// FEMMed.cpp : Defines the class behaviors for the application.
//

#include "stdafx.h"
#include "FEMMed.h"

#include "MainFrm.h"
#include "ChildFrm.h"
#include "FEMMedDoc.h"
#include "FEMMedView.h"

#include "Hyperlink.h"
#include <afxadv.h>  // Required for CRecentFileList m_pRecentFileList
                     // in OnOpenRecentFile().

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// Global variables.
BOOL g_bGlobalOVR = FALSE;  // OVR/INS status.
int g_nUntitledCount = 0;   // For use in CFEMMedDoc::SetTitle()
// Preferences.
BOOL g_bAutoCompletion     = TRUE;
BOOL g_bAutoIndentation    = TRUE;
BOOL g_bIntelliTip         = TRUE;
BOOL g_bSyntaxHighlighting = TRUE;
Indent g_enumIndent        = TWOSPACES;
// FEMM Lua functions and Lua keywords.
CFEMMedCObArray g_FEMMLuaFunctions;
CFEMMedCObArray g_LuaKeywords;
// Styles.
CSymbolStyle g_symComment;
CSymbolStyle g_symFunction;
CSymbolStyle g_symKeyword;
CSymbolStyle g_symNumber;
CSymbolStyle g_symPunctuation;
CSymbolStyle g_symString;
// Find/Replace global data.
CString g_strFind      = "";
CString g_strReplace   = "";
BOOL g_bMatchCase      = FALSE;
BOOL g_bMatchWholeWord = FALSE;

/////////////////////////////////////////////////////////////////////////////
// CFEMMedApp

BEGIN_MESSAGE_MAP(CFEMMedApp, CWinApp)
	//{{AFX_MSG_MAP(CFEMMedApp)
	ON_COMMAND(ID_APP_ABOUT, OnAppAbout)
	ON_COMMAND_EX_RANGE(ID_FILE_MRU_FILE1, ID_FILE_MRU_FILE16, OnOpenRecentFile)
	ON_COMMAND(ID_PREFERENCES, OnPreferences)
	//}}AFX_MSG_MAP
	// Standard file based document commands
	ON_COMMAND(ID_FILE_NEW, CWinApp::OnFileNew)
	ON_COMMAND(ID_FILE_OPEN, CWinApp::OnFileOpen)
	// Standard print setup command
	ON_COMMAND(ID_FILE_PRINT_SETUP, CWinApp::OnFilePrintSetup)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CFEMMedApp construction

CFEMMedApp::CFEMMedApp()
{
	// INI file should be located in same directory as executable.
	m_strFilename = __argv[0];
	m_strFilename = m_strFilename.Left(m_strFilename.ReverseFind('\\') + 1) + "FEMMed.ini";

	// Set defaults.
	::g_symComment.SetStyle(RGB(0,192,0), TRUE, FALSE, FALSE);
	::g_symFunction.SetStyle(RGB(0,0,0), TRUE, FALSE, FALSE);
	::g_symKeyword.SetStyle(RGB(0,0,255), FALSE, FALSE, FALSE);
	::g_symNumber.SetStyle(RGB(255,0,255), FALSE, FALSE, FALSE);
	::g_symPunctuation.SetStyle(RGB(128,128,128), TRUE, FALSE, FALSE);
	::g_symString.SetStyle(RGB(255,0,0), FALSE, FALSE, FALSE);

	m_strDefaultHeader =	"-- This is the INI file for use with FEMMed. Please keep it in the same\r\n" \
							"-- directory as the FEMMed executable (FEMMed.exe).\r\n" \
							"--\r\n" \
							"-- Single line comments can be added to this header the same as Lua.\r\n" \
							"--\r\n" \
							"-- If this file doesn't exist, then FEMMed will create it automatically\r\n" \
							"-- next time it is run.\r\n" \
							"--\r\n" \
							"-- Do not alter this file whilst FEMMed is running as all changes will\r\n" \
							"-- be overwritten when it is closed down.\r\n" \
							"--\r\n" \
							"-- Comments below this line will be appended to this header when FEMMed closes.\r\n";

	m_strDefaultLuaKeywords =	"and\r\n" \
								"break\r\n" \
								"do indent\r\n" \
								"else indent outdent\r\n" \
								"elseif indent outdent\r\n" \
								"end outdent\r\n" \
								"false\r\n" \
								"for indent\r\n" \
								"function indent\r\n" \
								"if indent\r\n" \
								"in\r\n" \
								"local\r\n" \
								"nil\r\n" \
								"not\r\n" \
								"or\r\n" \
								"repeat indent\r\n" \
								"return\r\n" \
								"then\r\n" \
								"true\r\n" \
								"until outdent\r\n" \
								"while indent\r\n";

	m_strDefaultFEMMLuaFunctions =	"clearconsole\r\n" \
									"create doctype\r\n" \
									"ei_addarc x1 y1 x2 y2 angle maxseg\r\n" \
									"ei_addblocklabel x y\r\n" \
									"ei_addboundprop \"boundpropname\" vs qs c0 c1 bdryformat\r\n" \
									"ei_addconductorprop \"conductorname\" vc qc conductortype\r\n" \
									"ei_addmaterial \"materialname\" ex ey qv\r\n" \
									"ei_addnode x y\r\n" \
									"ei_addpointprop \"pointpropname\" vp qp\r\n" \
									"ei_addsegment x1 y1 x2 y2\r\n" \
									"ei_analyze flag\r\n" \
									"ei_clearselected\r\n" \
									"ei_close\r\n" \
									"ei_copyrotate bx by angle copies (editaction)\r\n" \
									"ei_copytranslate dx dy copies (editaction)\r\n" \
									"ei_createmesh\r\n" \
									"ei_deleteboundprop \"boundpropname\"\r\n" \
									"ei_deleteconductor \"conductorname\"\r\n" \
									"ei_deletematerial \"materialname\"\r\n" \
									"ei_deletepointprop \"pointpropname\"\r\n" \
									"ei_deleteselected\r\n" \
									"ei_deleteselectedarcsegments\r\n" \
									"ei_deleteselectedlabels\r\n" \
									"ei_deleteselectednodes\r\n" \
									"ei_deleteselectedsegments\r\n" \
									"ei_gridsnap \"flag\"\r\n" \
									"ei_hidegrid\r\n" \
									"ei_loadsolution\r\n" \
									"ei_mirror x1 y1 x2 y2 (editaction)\r\n" \
									"ei_modifyboundprop \"bdryname\" propnum value\r\n" \
									"ei_modifyconductorprop \"conductorname\" propnum value\r\n" \
									"ei_modifymaterial \"blockname\" propnum value\r\n" \
									"ei_modifypointprop \"pointname\" propnum value\r\n" \
									"ei_moverotate bx by shiftangle (editaction)\r\n" \
									"ei_movetranslate dx dy (editaction)\r\n" \
									"ei_probdef units type precision (depth) (minangle)\r\n" \
									"ei_purgemesh\r\n" \
									"ei_readdxf \"filename\"\r\n" \
									"ei_refreshview\r\n" \
									"ei_saveas \"filename\"\r\n" \
									"ei_savebitmap \"filename\"\r\n" \
									"ei_savemetafile \"filename\"\r\n" \
									"ei_scale bx by scalefactor (editaction)\r\n" \
									"ei_selectarcsegment x y\r\n" \
									"ei_selectgroup n\r\n" \
									"ei_selectlabel x y\r\n" \
									"ei_selectnode x y\r\n" \
									"ei_selectsegment x y\r\n" \
									"ei_setarcsegmentprop maxsegdeg \"propname\" hide group \"inconductor\"\r\n" \
									"ei_setblockprop \"blockname\" automesh meshsize group\r\n" \
									"ei_seteditmode editmode\r\n" \
									"ei_setfocus \"documentname\"\r\n" \
									"ei_setgrid density \"type\"\r\n" \
									"ei_setnodeprop \"propname\" groupno \"inconductor\"\r\n" \
									"ei_setsegmentprop \"propname\" elementsize automesh hide group \"inconductor\"\r\n" \
									"ei_showgrid\r\n" \
									"ei_showmesh\r\n" \
									"ei_shownames flag\r\n" \
									"ei_zoom x1 y1 x2 y2\r\n" \
									"ei_zoomin\r\n" \
									"ei_zoomnatural\r\n" \
									"ei_zoomout\r\n" \
									"eo_addcontour x y\r\n" \
									"eo_bendcontour angle anglestep\r\n" \
									"eo_blockintegral type\r\n" \
									"eo_clearblock\r\n" \
									"eo_clearcontour\r\n" \
									"eo_close\r\n" \
									"eo_getconductorproperties \"conductor\"\r\n" \
									"eo_getpointvalues x y\r\n" \
									"eo_getprobleminfo\r\n" \
									"eo_gridsnap \"flag\"\r\n" \
									"eo_groupselectblock n\r\n" \
									"eo_hidecontourplot\r\n" \
									"eo_hidedensityplot\r\n" \
									"eo_hidegrid\r\n" \
									"eo_hidemesh\r\n" \
									"eo_hidepoints\r\n" \
									"eo_lineintegral type\r\n" \
									"eo_makeplot plottype numpoints filename fileformat\r\n" \
									"eo_refreshview\r\n" \
									"eo_reload\r\n" \
									"eo_savebitmap \"filename\"\r\n" \
									"eo_savemetafile \"filename\"\r\n" \
									"eo_selectblock x y\r\n" \
									"eo_selectconductor \"name\"\r\n" \
									"eo_selectpoint x y\r\n" \
									"eo_seteditmode mode\r\n" \
									"eo_setgrid density \"type\"\r\n" \
									"eo_showcontourplot numcontours lower_v upper_v\r\n" \
									"eo_showdensityplot legend gscale upper_d lower_d type\r\n" \
									"eo_showgrid\r\n" \
									"eo_showmesh\r\n" \
									"eo_shownames flag\r\n" \
									"eo_showpoints\r\n" \
									"eo_showvectorplot type scalefactor\r\n" \
									"eo_smooth \"flag\"\r\n" \
									"eo_zoom x1 y1 x2 y2\r\n" \
									"eo_zoomin\r\n" \
									"eo_zoomnatural\r\n" \
									"eo_zoomout\r\n" \
									"hideconsole\r\n" \
									"hidepointprops\r\n" \
									"messagebox \"message\"\r\n" \
									"mi_addarc x1 y1 x2 y2 angle maxseg\r\n" \
									"mi_addbhpoint \"blockname\" b h\r\n" \
									"mi_addblocklabel x y\r\n" \
									"mi_addboundprop \"propname\" a0 a1 a2 phi mu sig c0 c1 bdryformat\r\n" \
									"mi_addmaterial \"materialname\" mu_x mu_y h_c jr ji cduct lam_d phi_hmax lam_fill lamtype phi_hx phi_hy\r\n" \
									"mi_addnode x y\r\n" \
									"mi_addpointprop \"pointpropname\" a_re a_im j_re j_im\r\n" \
									"mi_addsegment x1 y1 x2 y2\r\n" \
									"mi_analyze flag\r\n" \
									"mi_clearbhpoints \"blockname\"\r\n" \
									"mi_clearselected\r\n" \
									"mi_close\r\n" \
									"mi_copyrotate bx by angle copies (editaction)\r\n" \
									"mi_copytranslate dx dy copies (editaction)\r\n" \
									"mi_createmesh\r\n" \
									"mi_deleteboundprop \"propname\"\r\n" \
									"mi_deletecircuit \"circuitname\"\r\n" \
									"mi_deletematerial \"materialname\"\r\n" \
									"mi_deletepointprop \"pointpropname\"\r\n" \
									"mi_deleteselected\r\n" \
									"mi_deleteselectedarcsegments\r\n" \
									"mi_deleteselectedlabels\r\n" \
									"mi_deleteselectednodes\r\n" \
									"mi_deleteselectedsegments\r\n" \
									"mi_grid_snap \"flag\"\r\n" \
									"mi_hidegrid\r\n" \
									"mi_loadsolution\r\n" \
									"mi_mirror x1 y1 x2 y2 (editaction)\r\n" \
									"mi_modifyboundprop \"bdryname\" propnum value\r\n" \
									"mi_modifycircprop \"circname\" propnum value\r\n" \
									"mi_modifymaterial \"blockname\" propnum value\r\n" \
									"mi_modifypointprop \"pointname\" propnum value\r\n" \
									"mi_moverotate bx by shiftangle (editaction)\r\n" \
									"mi_movetranslate dx dy (editaction)\r\n" \
									"mi_probdef frequency units type precision (depth) (minangle)\r\n" \
									"mi_purgemesh\r\n" \
									"mi_readdxf \"filename\"\r\n" \
									"mi_refreshview\r\n" \
									"mi_saveas \"filename\"\r\n" \
									"mi_savebitmap \"filename\"\r\n" \
									"mi_savemetafile \"filename\"\r\n" \
									"mi_scale bx by scalefactor (editaction)\r\n" \
									"mi_selectarcsegment x y\r\n" \
									"mi_selectgroup n\r\n" \
									"mi_selectlabel x y\r\n" \
									"mi_selectnode x y\r\n" \
									"mi_selectsegment x y\r\n" \
									"mi_setarcsegmentprop maxsegdeg \"propname\" hide group\r\n" \
									"mi_setblockprop \"blockname\" automesh meshsize \"incircuit\" magdirection group turns\r\n" \
									"mi_seteditmode editmode\r\n" \
									"mi_setfocus \"documentname\"\r\n" \
									"mi_setgrid density \"type\"\r\n" \
									"mi_setnodeprop \"propname\" groupno\r\n" \
									"mi_setsegmentprop \"propname\" elementsize automesh hide group\r\n" \
									"mi_showgrid\r\n" \
									"mi_showmesh\r\n" \
									"mi_shownames flag\r\n" \
									"mi_zoom x1 y1 x2 y2\r\n" \
									"mi_zoomin\r\n" \
									"mi_zoomnatural\r\n" \
									"mi_zoomout\r\n" \
									"mo_addcontour x y\r\n" \
									"mo_bendcontour angle anglestep\r\n" \
									"mo_blockintegral type\r\n" \
									"mo_clearblock\r\n" \
									"mo_clearcontour\r\n" \
									"mo_close\r\n" \
									"mo_getcircuitproperties \"circuit\"\r\n" \
									"mo_getpointvalues x y\r\n" \
									"mo_getprobleminfo\r\n" \
									"mo_grid_snap \"flag\"\r\n" \
									"mo_groupselectblock n\r\n" \
									"mo_hidecontourplot\r\n" \
									"mo_hidedensityplot\r\n" \
									"mo_hidegrid\r\n" \
									"mo_hidemesh\r\n" \
									"mo_hidepoints\r\n" \
									"mo_lineintegral type\r\n" \
									"mo_makeplot plottype numpoints filename fileformat\r\n" \
									"mo_refreshview\r\n" \
									"mo_reload\r\n" \
									"mo_savebitmap \"filename\"\r\n" \
									"mo_savemetafile \"filename\"\r\n" \
									"mo_selectblock x y\r\n" \
									"mo_selectpoint x y\r\n" \
									"mo_seteditmode mode\r\n" \
									"mo_setgrid density \"type\"\r\n" \
									"mo_showcontourplot numcontours lower_a upper_a type\r\n" \
									"mo_showdensityplot legend gscale upper_b lower_b type\r\n" \
									"mo_showgrid\r\n" \
									"mo_showmesh\r\n" \
									"mo_shownames flag\r\n" \
									"mo_showpoints\r\n" \
									"mo_smooth \"flag\"\r\n" \
									"mo_zoom x1 y1 x2 y2\r\n" \
									"mo_zoomin\r\n" \
									"mo_zoomnatural\r\n" \
									"mo_zoomout\r\n" \
									"open \"filename\"\r\n" \
									"pause\r\n" \
									"print\r\n" \
									"prompt \"message\"\r\n" \
									"quit\r\n" \
									"showconsole\r\n" \
									"showpointprops\r\n";
}

/////////////////////////////////////////////////////////////////////////////
// CFEMMedApp destruction

CFEMMedApp::~CFEMMedApp()
{
	// De-allocate memory from CFEMMedCObArrays.
	int i=0;
	while(i < ::g_FEMMLuaFunctions.GetSize())
	{
		delete ::g_FEMMLuaFunctions.GetAt(i++);
	}

	i=0;
	while(i < ::g_LuaKeywords.GetSize())
	{
		delete ::g_LuaKeywords.GetAt(i++);
	}
}

/////////////////////////////////////////////////////////////////////////////
// The one and only CFEMMedApp object

CFEMMedApp theApp;

/////////////////////////////////////////////////////////////////////////////
// CFEMMedApp initialization

BOOL CFEMMedApp::InitInstance()
{
	// Initialize OLE libraries
	if (!AfxOleInit())
	{
		AfxMessageBox(IDP_OLE_INIT_FAILED);
		return FALSE;
	}

	// Standard initialization
	// If you are not using these features and wish to reduce the size
	//  of your final executable, you should remove from the following
	//  the specific initialization routines you do not need.

#ifdef _AFXDLL
	Enable3dControls();			// Call this when using MFC in a shared DLL
#else
	Enable3dControlsStatic();	// Call this when linking to MFC statically
#endif

	// Change the registry key under which our settings are stored.
	// TODO: You should modify this string to be something appropriate
	// such as the name of your company or organization.
	SetRegistryKey(_T("Local AppWizard-Generated Applications"));

	LoadStdProfileSettings();  // Load standard INI file options (including MRU)

	// Register the application's document templates.  Document templates
	//  serve as the connection between documents, frame windows and views.

	CMultiDocTemplate* pDocTemplate;
	pDocTemplate = new CMultiDocTemplate(
		IDR_FEMMEDTYPE,
		RUNTIME_CLASS(CFEMMedDoc),
		RUNTIME_CLASS(CChildFrame), // custom MDI child frame
		RUNTIME_CLASS(CFEMMedView));
	pDocTemplate->SetContainerInfo(IDR_FEMMEDTYPE_CNTR_IP);
	AddDocTemplate(pDocTemplate);

	// create main MDI Frame window
	CMainFrame* pMainFrame = new CMainFrame;
	if (!pMainFrame->LoadFrame(IDR_MAINFRAME))
		return FALSE;
	m_pMainWnd = pMainFrame;

	// Enable DDE Execute open
	EnableShellOpen();
	RegisterShellFileTypes(TRUE);

	// Parse command line for standard shell commands, DDE, file open
	CCommandLineInfo cmdInfo;
	ParseCommandLine(cmdInfo);

	// Dispatch commands specified on the command line
	if (!ProcessShellCommand(cmdInfo))
		return FALSE;

	// The main window has been initialized, so show and update it.
	pMainFrame->ShowWindow(m_nCmdShow);
	pMainFrame->UpdateWindow();

	LoadINIFile();

	return TRUE;
}

// Load the INI file and set the program's data.
void CFEMMedApp::LoadINIFile()
{
	try
	{
		CStdioFile file(m_strFilename, CFile::modeRead);
		
		CString strLine, strOptions, strCommentStyle, strFunctionStyle,
		strKeywordStyle, strNumberStyle, strPunctuationStyle, strStringStyle,
		strLuaKeywords, strFEMMLuaFunctions;

		int nFoundComment = -1;
		BOOL bAddToOptions, bAddToCommentStyle, bAddToFunctionStyle, bAddToKeywordStyle,
			bAddToNumberStyle, bAddToPunctuationStyle, bAddToStringStyle, bAddToLuaKeywords,
			bAddToLuaFEMMFunctions;

		bAddToOptions = bAddToCommentStyle = bAddToFunctionStyle = bAddToKeywordStyle = 
			bAddToNumberStyle = bAddToPunctuationStyle= bAddToStringStyle =
			bAddToLuaKeywords = bAddToLuaFEMMFunctions = FALSE;

		// Break file string into smalled components.
		while(file.ReadString(strLine))
		{
			// Remove comments.
			nFoundComment = strLine.Find("--");
			if(nFoundComment != -1)
			{
				// Add comments to the header and delete them.
				m_strHeader += 
					strLine.Mid(nFoundComment, strLine.GetLength() - nFoundComment) + "\n";

				strLine.Delete(nFoundComment, strLine.GetLength() - nFoundComment);
			}

			if(!strLine.IsEmpty() && strLine != "")
			{

				strLine.MakeLower();

				if(strLine.Find("[/options]") >= 0)
					bAddToOptions = FALSE;
				else if(strLine.Find("[/comment_style]") >= 0)
					bAddToCommentStyle = FALSE;
				else if(strLine.Find("[/function_style]") >= 0)
					bAddToFunctionStyle = FALSE;
				else if(strLine.Find("[/keyword_style]") >= 0)
					bAddToKeywordStyle = FALSE;
				else if(strLine.Find("[/number_style]") >= 0)
					bAddToNumberStyle = FALSE;
				else if(strLine.Find("[/punctuation_style]") >= 0)
					bAddToPunctuationStyle = FALSE;
				else if(strLine.Find("[/string_style]") >= 0)
					bAddToStringStyle = FALSE;
				else if(strLine.Find("[/lua_keywords]") >= 0)
					bAddToLuaKeywords = FALSE;
				else if(strLine.Find("[/femm_lua_functions]") >= 0)
					bAddToLuaFEMMFunctions = FALSE;

				if(bAddToOptions)
					strOptions += strLine + "\r\n";
				else if(bAddToCommentStyle)
					strCommentStyle += strLine + "\r\n";
				else if(bAddToFunctionStyle)
					strFunctionStyle += strLine + "\r\n";
				else if(bAddToKeywordStyle)
					strKeywordStyle += strLine + "\r\n";
				else if(bAddToNumberStyle)
					strNumberStyle += strLine + "\r\n";
				else if(bAddToPunctuationStyle)
					strPunctuationStyle += strLine + "\r\n";
				else if(bAddToStringStyle)
					strStringStyle += strLine + "\r\n";
				else if(bAddToLuaKeywords)
					strLuaKeywords += strLine + "\r\n";
				else if(bAddToLuaFEMMFunctions)
					strFEMMLuaFunctions += strLine + "\r\n";

				if(strLine.Find("[options]") >= 0)
					bAddToOptions = TRUE;
				else if(strLine.Find("[comment_style]") >= 0)
					bAddToCommentStyle = TRUE;
				else if(strLine.Find("[function_style]") >= 0)
					bAddToFunctionStyle = TRUE;
				else if(strLine.Find("[keyword_style]") >= 0)
					bAddToKeywordStyle = TRUE;
				else if(strLine.Find("[number_style]") >= 0)
					bAddToNumberStyle = TRUE;
				else if(strLine.Find("[punctuation_style]") >= 0)
					bAddToPunctuationStyle = TRUE;
				else if(strLine.Find("[string_style]") >= 0)
					bAddToStringStyle = TRUE;
				else if(strLine.Find("[lua_keywords]") >= 0)
					bAddToLuaKeywords = TRUE;
				else if(strLine.Find("[femm_lua_functions]") >= 0)
					bAddToLuaFEMMFunctions = TRUE;
			}

			// Reset nFoundComment.
			nFoundComment = -1;
		}

		// Set options.
		::g_bAutoCompletion = ExtractBOOL(strOptions, "autocompletion");
		::g_bAutoIndentation = ExtractBOOL(strOptions, "indentation");
		::g_enumIndent = ExtractEnum(strOptions, "indent");
		::g_bIntelliTip = ExtractBOOL(strOptions, "intellitip");
		::g_bSyntaxHighlighting = ExtractBOOL(strOptions, "syntaxhighlighting");

		// Set symbol styles.
		SetStyleFromString(::g_symComment, strCommentStyle);
		SetStyleFromString(::g_symFunction, strFunctionStyle);
		SetStyleFromString(::g_symKeyword, strKeywordStyle);
		SetStyleFromString(::g_symNumber, strNumberStyle);
		SetStyleFromString(::g_symPunctuation, strPunctuationStyle);
		SetStyleFromString(::g_symString, strStringStyle);

		// Add FEMM Lua functions.
		::g_FEMMLuaFunctions.AddMultiLineString(strFEMMLuaFunctions);
		::g_FEMMLuaFunctions.Sort();

		// Add Lua keywords.
		::g_LuaKeywords.AddMultiLineString(strLuaKeywords);
		::g_LuaKeywords.Sort();
	}
	catch(CFileException* e)
	{
		if(e->m_cause == CFileException::fileNotFound)
		{
			MessageBox(this->m_pMainWnd->GetSafeHwnd(), "\"" + e->m_strFileName + "\""\
				+ "\n\nIf this is the first time the program has been run, igore" \
				" this message.\nDefaults will be saved when program closed.", \
				"FEMMed - Could not load INI file.", MB_OK | MB_ICONWARNING);
			
			// Default header.
			m_strHeader = m_strDefaultHeader;

			// Add Default FEMM Lua functions.
			::g_FEMMLuaFunctions.AddMultiLineString(m_strDefaultFEMMLuaFunctions);
			::g_FEMMLuaFunctions.Sort();

			// Add Default Lua keywords.
			::g_LuaKeywords.AddMultiLineString(m_strDefaultLuaKeywords);
			::g_LuaKeywords.Sort();
		}
		else
		{
			e->ReportError();
		}

		e->Delete();
	}
}

// Set CSymbolStyle members from string.
void CFEMMedApp::SetStyleFromString(CSymbolStyle& symStyle, const CString str)
{
	if(str.IsEmpty())
		return;

	symStyle.SetStyle(
		ExtractClr(str),
		ExtractBOOL(str, "bold"),
		ExtractBOOL(str, "italic"),
		ExtractBOOL(str, "underline")
	);
}

// Extract BOOL from string.
BOOL CFEMMedApp::ExtractBOOL(CString str, CString strToExtract)
{
	// Remove spaces and tabs.
	str.Remove(' ');
	str.Remove('\t');

	strToExtract += "=";

	str = str.Right(str.GetLength() - (str.Find(strToExtract) + strToExtract.GetLength()));
	str = str.Left(str.Find("\r\n"));
	
	return str.Find("true") < 0 ? FALSE : TRUE;
}

// Extract COLORREFfrom string.
COLORREF CFEMMedApp::ExtractClr(CString str)
{
	// Remove spaces and tabs.
	str.Remove(' ');
	str.Remove('\t');

	str = str.Right(str.GetLength() - (str.Find("color=")+6));  // +6 to exclude "color=".
	str = str.Left(str.Find("\r\n"));

	int clrR, clrG, clrB;
	clrR = clrG = clrB = 0;

	clrR = atoi(str.Left(str.Find(',')));
	str = str.Right(str.GetLength() - str.Find(',') - 1);
	clrG = atoi(str.Left(str.Find(',')));
	str = str.Right(str.GetLength() - str.Find(',') - 1);
	clrB = atoi(str);
	
	return RGB(clrR, clrG, clrB);
}

// Extract enumerated type from string.
::Indent CFEMMedApp::ExtractEnum(CString str, CString strToExtract)
{
	// Remove spaces and tabs.
	str.Remove(' ');
	str.Remove('\t');

	strToExtract += "=";

	str = str.Right(str.GetLength() - (str.Find(strToExtract) + strToExtract.GetLength()));
	str = str.Left(str.Find("\r\n"));
	
	if(str == "tab")
		return ::TAB;
	else if(str == "fourspaces")
		return ::FOURSPACES;
	else  // if(str == "twospaces");
		return ::TWOSPACES;
}

/////////////////////////////////////////////////////////////////////////////
// CAboutDlg dialog used for App About

class CAboutDlg : public CDialog
{
public:
	CAboutDlg();

// Dialog Data
	//{{AFX_DATA(CAboutDlg)
	enum { IDD = IDD_ABOUTBOX };
	//}}AFX_DATA

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CAboutDlg)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

// Implementation
protected:
	//{{AFX_MSG(CAboutDlg)
		// No message handlers
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()

// Hyperlinks.
private:
	CHyperLink m_hlAutoCompletion;
	CHyperLink m_hlHyperLink;
	CHyperLink m_hlIntelliTip;
	CHyperLink m_hlMTI;
	CHyperLink m_hlSyntaxHighlighting;
};

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
	//{{AFX_DATA_INIT(CAboutDlg)
	//}}AFX_DATA_INIT
	m_hlAutoCompletion.SetURL(_T("http://www.codeproject.com/richedit/AutocomplectionCtrl.asp"));
	m_hlHyperLink.SetURL(_T("http://www.codeproject.com/miscctrl/hyperlink.asp"));
	m_hlIntelliTip.SetURL(_T("http://www.codeproject.com/miscctrl/intellitip.asp"));
	m_hlMTI.SetURL(_T("http://www.codeproject.com/docview/tabbedmdi.asp"));
	m_hlSyntaxHighlighting.SetURL(_T("http://www.codeguru.com/Cpp/controls/richedit/syntaxhilighting/article.php/c2387/"));
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAboutDlg)
	DDX_Control(pDX, IDHL_AUTOCOMPLETION, m_hlAutoCompletion);
	DDX_Control(pDX, IDHL_HYPERLINK, m_hlHyperLink);
	DDX_Control(pDX, IDHL_INTELLITIP, m_hlIntelliTip);
	DDX_Control(pDX, IDHL_MTI, m_hlMTI);
	DDX_Control(pDX, IDHL_SYNTAXHIGHLIGHTING, m_hlSyntaxHighlighting);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
	//{{AFX_MSG_MAP(CAboutDlg)
		// No message handlers
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CFEMMedApp message handlers

// Write INI file.

// Important to note that CStdioFile::WriteString() replaces any newline
// character with a CR/LF pair. So only '\n' is used.
int CFEMMedApp::ExitInstance()
{
	try
	{
		CStdioFile file(m_strFilename, CFile::modeWrite | CFile::modeCreate);

		// Write header.
		file.WriteString(m_strHeader);

		// Blank line.
		file.WriteString("\n");

		// Options.
		file.WriteString("[options]\n");
		WriteBOOL(file, "autocompletion", ::g_bAutoCompletion);
		WriteBOOL(file, "autoindentation", ::g_bAutoIndentation);
		switch(::g_enumIndent)
		{
		case ::TAB:
			file.WriteString("indent = TAB\n");
			break;
		case ::FOURSPACES:
			file.WriteString("indent = FOURSPACES\n");
			break;
		case ::TWOSPACES:
		default:
			file.WriteString("indent = TWOSPACES\n");
			break;
		}
		WriteBOOL(file, "intellitip", ::g_bIntelliTip);
		WriteBOOL(file, "syntaxhighlighting", ::g_bSyntaxHighlighting);
		file.WriteString("[/options]\n");

		// Blank line.
		file.WriteString("\n");

		// Comment style.
		file.WriteString("[comment_style]\n");
		WriteStyle(file, ::g_symComment);
		file.WriteString("[/comment_style]\n");

		// Blank line.
		file.WriteString("\n");

		// Function style.
		file.WriteString("[function_style]\n");
		WriteStyle(file, ::g_symFunction);
		file.WriteString("[/function_style]\n");
		
		// Blank line.
		file.WriteString("\n");

		// Keyword style.
		file.WriteString("[keyword_style]\n");
		WriteStyle(file, ::g_symKeyword);
		file.WriteString("[/keyword_style]\n");
		
		// Blank line.
		file.WriteString("\n");

		// Number style.
		file.WriteString("[number_style]\n");
		WriteStyle(file, ::g_symNumber);
		file.WriteString("[/number_style]\n");
		
		// Blank line.
		file.WriteString("\n");

		// Punctuation style.
		file.WriteString("[punctuation_style]\n");
		WriteStyle(file, ::g_symPunctuation);
		file.WriteString("[/punctuation_style]\n");
		
		// Blank line.
		file.WriteString("\n");

		// String style.
		file.WriteString("[string_style]\n");
		WriteStyle(file, ::g_symString);
		file.WriteString("[/string_style]\n");
		
		// Blank line.
		file.WriteString("\n");

		// Lua keywords.
		file.WriteString("[lua_keywords]\n");
		WriteFEMMedCObArray(file, ::g_LuaKeywords);
		file.WriteString("[/lua_keywords]\n");
		
		// Blank line.
		file.WriteString("\n");

		// FEMM Lua functions.
		file.WriteString("[femm_lua_functions]\n");
		WriteFEMMedCObArray(file, ::g_FEMMLuaFunctions);
		file.WriteString("[/femm_lua_functions]\n");
	}
	catch(CFileException* e)
	{
		e->ReportError();
		e->Delete();
	}

	return CWinApp::ExitInstance();
}

// Write a BOOL to the INI file.
void CFEMMedApp::WriteBOOL(CStdioFile& file, CString str, const BOOL b)
{
	if(b)
		str += " = TRUE\n";
	else
		str += " = FALSE\n";

	file.WriteString(str);
}

// Write a COLORREF to the INI file.
void CFEMMedApp::WriteCOLORREF(CStdioFile& file, const COLORREF clr)
{
	CString str;
	str.Format("color = %i, %i, %i\n", GetRValue(clr), GetGValue(clr), GetBValue(clr));

	file.WriteString(str);
}

// Write a CFEMMedCObArray object to the INI file.
void CFEMMedApp::WriteFEMMedCObArray(CStdioFile& file, const CFEMMedCObArray& FEMMedCObArray)
{
	for(int i=0; i<FEMMedCObArray.GetSize(); i++)
	{
		CStringArray* pSA = (CStringArray*)FEMMedCObArray.GetAt(i);
		for(int j=0; j<pSA->GetSize(); j++)
		{
			if(j<pSA->GetUpperBound())
				file.WriteString(pSA->GetAt(j) + " ");
			else
				file.WriteString(pSA->GetAt(j) + "\n");
		}
	}
}

// Write a CSymbolStyle object to the INI file.
void CFEMMedApp::WriteStyle(CStdioFile& file, const CSymbolStyle& symStyle)
{
	// Color.
	WriteCOLORREF(file, symStyle.GetColor());

	// Bold.
	WriteBOOL(file, "bold", symStyle.GetBold());

	// Italic.
	WriteBOOL(file, "italic", symStyle.GetItalic());

	// Underline.
	WriteBOOL(file, "underline", symStyle.GetUnderline());
}

// Cycle through all open views and hide the list box and IntelliTip if required,
// and reformat the view's syntax color highlighting.
void CFEMMedApp::FormatAllViews()
{
	if(m_pDocManager != NULL)
	{
		// Walk all templates.
		CDocTemplate* pTemplate;
		POSITION pos = CFEMMedApp::m_pDocManager->GetFirstDocTemplatePosition();
		while (pos != NULL)
		{
			pTemplate = CFEMMedApp::m_pDocManager->GetNextDocTemplate(pos);
			ASSERT(pTemplate);
			// Walk all documents in the template.
			POSITION pos2 = pTemplate->GetFirstDocPosition();
			while (pos2)
			{
				CDocument* pDoc = pTemplate->GetNextDoc(pos2);
				ASSERT(pDoc);
				// Walk all views for each document.
				POSITION pos3 = pDoc->GetFirstViewPosition();
				while(pos3)
				{
					CFEMMedView* pView = (CFEMMedView*)pDoc->GetNextView(pos3);
					ASSERT(pView);
					pView->FormatAll();

					if(!::g_bAutoCompletion)
					{
						pView->RemoveListbox();
					}
					if(!::g_bIntelliTip)
					{
						pView->m_IntelliTip.Hide();
					}
				}
			}
		}
	}
}

// App command to run the dialog
void CFEMMedApp::OnAppAbout()
{
	CAboutDlg aboutDlg;
	aboutDlg.DoModal();
}

// This function is used to check that a file exists before it is opened from the
// Most Recently Used (MRU) file list. It was found, without explanation, that if a
// file on the MRU had been deleted or moved, that the program would crash.
BOOL CFEMMedApp::OnOpenRecentFile(UINT nID)
{
	int nIndex = nID - ID_FILE_MRU_FILE1;
	CFileStatus fs;

	if(!CFile::GetStatus((*m_pRecentFileList)[nIndex], fs))
	{
		MessageBox(this->m_pMainWnd->GetSafeHwnd(), "\"" + (*m_pRecentFileList)[nIndex] \
				+ "\"" + "\n\nThe file will be removed from the most recently used " \
				"file list.", "FEMMed - Could not load file.", MB_OK | MB_ICONWARNING);

		m_pRecentFileList->Remove(nIndex);
	}
	else
	{
		OpenDocumentFile((*m_pRecentFileList)[nIndex]);
	}

	return TRUE;
}

// Preferences command to run the dialog.
void CFEMMedApp::OnPreferences() 
{
	CFEMMedPreferencesDlg prefDlg;
	prefDlg.DoModal();
}
