#include "stdafx.h"
#include "FEMMedCObArray.h"

IMPLEMENT_DYNAMIC(CFEMMedCObArray, CObArray)

// Call CObArray's constructor.
CFEMMedCObArray::CFEMMedCObArray() : CObArray()
{
}

// Return the maximum number of elements in any of the CStringArrays.
int CFEMMedCObArray::GetMaxParams()
{
	int max_elements = 0;

	for(int i=0; i<GetSize(); i++)
	{
		// GetSize() - 1 as first element is not parameter.
		if(((CStringArray*)GetAt(i))->GetSize() - 1 > max_elements)
			max_elements = ((CStringArray*)GetAt(i))->GetSize() - 1;
	}

	return max_elements;
}

// Add Lua FEMM functions.
void CFEMMedCObArray::AddMultiLineString(CString str)
{
	int nStart, nEnd, nIndex, nFound;
	nStart = nEnd = nIndex = nFound = 0;
	CString strTemp, strAdd;
	strTemp = strAdd = "";

	// Replace spaces and tabs with arbitrary character: %.
	str.Replace(' ', '%');
	str.Replace('\t', '%');

	while((nEnd = str.Find("\r\n", nStart)) != -1)
	{
		strTemp = str.Left(nEnd).Right(nEnd - nStart);
		if(!strTemp.IsEmpty())
		{
			Add(new CStringArray);
			do
			{
				if((nFound = strTemp.Find('%')) == -1)
				{
					strAdd = strTemp;
				}
				else
				{
					strAdd = strTemp.Left(nFound);
					strTemp.Delete(0, nFound + 1);
				}
				if(!strAdd.IsEmpty())
					((CStringArray*)GetAt(nIndex))->Add(strAdd);
			} while(nFound != -1);
			
			nIndex++;
		}

		nStart = nEnd + 2;
	}
}

// Sort the array.
void CFEMMedCObArray::Sort()
{
	BOOL bNotDone = TRUE;

	while (bNotDone)	// Keep sorting until complete.
	{					// i.e. no more SetAt operations in CompareAndSwap().
		bNotDone = FALSE;
		for(int nPos=0; nPos<GetUpperBound(); nPos++)
		{
			bNotDone |= CompareAndSwap(nPos);
		}
	}
}

/*
 * This is the algorithm used by Sort() to arrange the objects correctly.
 * This implementation is specific to a CObArray derived class that contains only
 * CStringArray objects and will not work for other types.
 *
 * It performs a case sensitive comparison on the first CString object in each
 * CStringArray with the first CString object in the following CStringArray. If the
 * first CString object is greater, then the two CStringArrays are swapped (i.e.
 * the first is moved down the CObArray derived object and the second moved up).
 */
BOOL CFEMMedCObArray::CompareAndSwap(const int nPos)
{
	int nPosFirst = nPos;
	int nPosNext  = nPos + 1;

	if(((CStringArray*)GetAt(nPosFirst))->GetAt(0).Compare(
		((CStringArray*)GetAt(nPosNext))->GetAt(0)) > 0)
	{
		CStringArray* tempCObj = (CStringArray*)GetAt(nPosFirst);
		SetAt(nPosFirst, GetAt(nPosNext));
		SetAt(nPosNext, tempCObj);

		return TRUE;
	}

	return FALSE;
}