// FEMMedView.cpp : implementation of the CFEMMedView class
//

#include "stdafx.h"
#include "FEMMed.h"

#include "FEMMedDoc.h"
#include "CntrItem.h"
#include "FEMMedView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// Required for use of RichEdit >= 2.0
#define DEFINE_GUIDXXX(name, l, w1, w2, b1, b2, b3, b4, b5, b6, b7, b8) \
	EXTERN_C const GUID CDECL name = { l, w1, w2, { b1, b2,  b3,  b4,  b5,  b6,  b7,  b8 } }
DEFINE_GUIDXXX(IID_ITextDocument, 0x8CC497C0, 0xA1DF, 0x11CE, 0x80, 0x98, 0x00,\
	0xAA, 0x00,0x47,0xBE,0x5D);

// Globals - see FEMMed.h and FEMMed.cpp.
extern BOOL g_bGlobalOVR;
extern BOOL g_bAutoCompletion;
extern BOOL g_bAutoIndentation;
extern BOOL g_bIntelliTip;
extern BOOL g_bSyntaxHighlighting;
extern Indent g_enumIndent;
extern CFEMMedCObArray g_FEMMLuaFunctions;
extern CFEMMedCObArray g_LuaKeywords;
extern CSymbolStyle g_symComment;
extern CSymbolStyle g_symFunction;
extern CSymbolStyle g_symKeyword;
extern CSymbolStyle g_symNumber;
extern CSymbolStyle g_symPunctuation;
extern CSymbolStyle g_symString;
extern CString g_strFind;
extern CString g_strReplace;
extern BOOL g_bMatchCase;
extern BOOL g_bMatchWholeWord;

/////////////////////////////////////////////////////////////////////////////
// CFEMMedView

// Register the message for the Find/Replace dialog.
const UINT CFEMMedView::m_FindDialogMessage = RegisterWindowMessage(FINDMSGSTRING);

IMPLEMENT_DYNCREATE(CFEMMedView, CRichEditView)

BEGIN_MESSAGE_MAP(CFEMMedView, CRichEditView)
	//{{AFX_MSG_MAP(CFEMMedView)
	ON_WM_DESTROY()
	ON_WM_CREATE()
	ON_CONTROL_REFLECT(EN_CHANGE, OnChange)
	ON_COMMAND(ID_EDIT_PASTE, OnEditPaste)
	ON_UPDATE_COMMAND_UI(ID_EDIT_PASTE, OnUpdateEditPaste)
	ON_WM_RBUTTONDOWN()
	ON_WM_CHAR()
	ON_WM_KEYDOWN()
	ON_WM_LBUTTONDOWN()
	ON_COMMAND(ID_EDIT_UNDO, OnEditUndo)
	ON_COMMAND(ID_EDIT_REDO, OnEditRedo)
	ON_UPDATE_COMMAND_UI(ID_EDIT_REDO, OnUpdateEditRedo)
	ON_COMMAND(ID_EDIT_FIND, OnEditFind)
	ON_COMMAND(ID_EDIT_REPLACE, OnEditReplace)
	ON_COMMAND(ID_EDIT_FINDNEXT, OnEditFindNext)
	ON_UPDATE_COMMAND_UI(ID_EDIT_FINDNEXT, OnUpdateEditFindNext)
	ON_UPDATE_COMMAND_UI(ID_POSITION_INDICATOR, OnUpdatePositionIndicator)
	//}}AFX_MSG_MAP
	// Standard printing commands
	ON_COMMAND(ID_FILE_PRINT, CRichEditView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, CRichEditView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, CRichEditView::OnFilePrintPreview)
	ON_NOTIFY_REFLECT(EN_PROTECTED, OnProtected)
	ON_NOTIFY_REFLECT(EN_SELCHANGE, OnSelChange)
	ON_MESSAGE(WM_SETTEXT, OnSetText)
	// FINDMSGSTRING registered message
	ON_REGISTERED_MESSAGE(m_FindDialogMessage, OnFindDialogMessage)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CFEMMedView construction/destruction

CFEMMedView::CFEMMedView() : m_REC(GetRichEditCtrl())
{
	// Use RichEdit >= 2.0
	m_strClass = _T(RICHEDIT_CLASSA);

	// Initialize pointer to ITextDocument interface.
	// Necessary to prevent illegal calls of ResumeRedo(), SuspendRedo()
	// and Release().
	m_pTextDoc = NULL;

	// OVR state.
	m_bLocalOVR = FALSE;
	m_bInToggleOVR = FALSE;

	// Has OnInitialUpdate() completed?
	m_bOnInitialUpdate = FALSE;

	// If freeze count is non-zero, screen updating is disabled.
	m_lFreezeCount = 0;

	// Character height, in twips (1/1440 of an inch, or 1/20 of a printer's point).
	m_lYHeight = 200;

	// Initialize pointer to Find/Replace dialog.
	m_pFindDialog = NULL;

	// Initialize auto-completion member.
	m_bAutoCompleted = FALSE;

	// Initialize auto-indent members.
	m_strIndent = "  ";
	m_nIndent = 0;
	m_bOutdent = FALSE;

	// Initialize syntax highlighting members.
	m_bInForcedChange = FALSE;
	m_changeType = ctUndo;
	m_crOldSel.cpMin = m_crOldSel.cpMax = 0;
}

CFEMMedView::~CFEMMedView()
{
	if(m_pTextDoc != NULL)
		m_pTextDoc->Release();

	m_IntelliTip.DestroyWindow();
}

void CFEMMedView::OnInitialUpdate()
{
	CRichEditView::OnInitialUpdate();

	// Prevent drag and drop files becoming embedded.
	m_REC.DragAcceptFiles(FALSE);

	// Initialize the pointer to ITextDocument: m_pTextDoc.
	IRichEditOle* ire = m_REC.GetIRichEditOle();
	if((ire == NULL) || (ire->QueryInterface(IID_ITextDocument, (void**)&m_pTextDoc) != S_OK)
		|| (m_pTextDoc == NULL))
	{
		if(ire != NULL)
			m_pTextDoc = NULL;
	}
	ire->Release();

	// Create IntelliTip.
	m_IntelliTip.Create(this);

	// Turn off word wrap.
	m_nWordWrap = WrapNone;
	WrapChanged();

	// Set the printing margins (720 twips = 1/2 inch).
	SetMargins(CRect(720, 720, 720, 1080));

	// OnInitialUpdate() complete, FormatAll() in case file has been loaded.
	m_bOnInitialUpdate = TRUE;
	FormatAll();
}

/////////////////////////////////////////////////////////////////////////////
// CFEMMedView printing

BOOL CFEMMedView::OnPreparePrinting(CPrintInfo* pInfo)
{
	pInfo->SetMaxPage(1);
	return DoPreparePrinting(pInfo);
}

void CFEMMedView::OnPrint(CDC* pDC, CPrintInfo* pInfo) 
{
	ASSERT_VALID(this);
	ASSERT_VALID(pDC);
	ASSERT(pInfo != NULL);
	ASSERT(pInfo->m_bContinuePrinting);

	UINT nCurPage = pInfo->m_nCurPage;
	ASSERT(nCurPage <= (UINT)m_aPageStart.GetSize());
	long lIndex = (long)m_aPageStart[nCurPage - 1];

	// Print as much as possible in the current page.
	lIndex = PrintPage(pDC, lIndex, 0xFFFFFFFF);

	// lIndex got above doesn't take "return" into account.
	// The following codes adds to lIndex when there are any new lines
	long lLength = GetTextLength();
	CString strRich;
	m_REC.GetWindowText(strRich);
	long lStart = 0;
	while(lStart != -1)
	{
		lStart = strRich.Find("\r\n", lStart);
		if(lStart != -1)
		{
			lLength--;
			lStart++;
		}

	}

	// Fix to error that might cause non-stop printing?
	lLength -= 10000;
	
	if (lIndex >= lLength)  // GetTextLength()
	{
		pInfo->SetMaxPage(nCurPage);
	}

	// Update pagination information for page just printed.
	if(nCurPage == (UINT)m_aPageStart.GetSize())
	{
		if(lIndex < lLength)  // GetTextLength()
			m_aPageStart.Add(lIndex);
	}
	else
	{
		ASSERT(nCurPage+1 <= (UINT)m_aPageStart.GetSize());
		ASSERT(lIndex == (long)m_aPageStart[nCurPage]);
	}
}

void CFEMMedView::OnDestroy()
{
	// Deactivate the item on destruction; this is important
	// when a splitter view is being used.
   CRichEditView::OnDestroy();
   COleClientItem* pActiveItem = GetDocument()->GetInPlaceActiveItem(this);
   if(pActiveItem != NULL && pActiveItem->GetActiveView() == this)
   {
      pActiveItem->Deactivate();
      ASSERT(GetDocument()->GetInPlaceActiveItem(this) == NULL);
   }
}


/////////////////////////////////////////////////////////////////////////////
// CFEMMedView diagnostics

#ifdef _DEBUG
void CFEMMedView::AssertValid() const
{
	CRichEditView::AssertValid();
}

void CFEMMedView::Dump(CDumpContext& dc) const
{
	CRichEditView::Dump(dc);
}

CFEMMedDoc* CFEMMedView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CFEMMedDoc)));
	return (CFEMMedDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CFEMMedView message handlers

void CFEMMedView::OnEditPaste() 
{
	// Paste plain text without formatting
	// View is sometimes scrolled is screen update not disabled.
	DisableScreenUpdate();
	m_REC.PasteSpecial(CF_TEXT);
	EnableScreenUpdate();
}

void CFEMMedView::OnUpdateEditPaste(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(::IsClipboardFormatAvailable(CF_TEXT));
}

HRESULT CFEMMedView::QueryAcceptData(LPDATAOBJECT lpDataObj,
		CLIPFORMAT FAR * lpcfFormat, DWORD dwReco, BOOL bReally, HGLOBAL hMetaFile)
{
	HRESULT result;
	COleDataObject oleDataObj;

	// Assume we won't be accepting the drag nor drop.
	result = S_FALSE;

	if(lpcfFormat)
	{
		// Cast the object to an ole source.
		oleDataObj.Attach(lpDataObj, FALSE);

		// If plain text is available.
		if(oleDataObj.IsDataAvailable(CF_TEXT))
		{
			// Indicate Accept the drop.
			result = S_OK;

			// If data dropped.
			if(bReally)
			{
				// force the drop to be in text format
				*lpcfFormat = (CLIPFORMAT)CF_TEXT;
			}
		}

		// Else plain text isn't available
		else
		{
			// If user is dragging the file around.
			if(oleDataObj.IsDataAvailable(CF_HDROP))
			{
				// Indicate accept the drop.
				result = S_OK;

				// If data dropped.
				if(bReally)
				{
					// Bring application's window to the front.
					AfxGetMainWnd()->SetForegroundWindow();

					// Force the drop to be in plain text format
					*lpcfFormat = (CLIPFORMAT)CF_TEXT;

					STGMEDIUM stgmedium;
					HGLOBAL hGlobal;

					oleDataObj.GetData((CLIPFORMAT) CF_HDROP, &stgmedium);
					if(stgmedium.tymed == TYMED_HGLOBAL)
						hGlobal = stgmedium.hGlobal;
					
					CString str;

					UINT nLength = DragQueryFile((HDROP)hGlobal, 0, NULL, 0);
					DragQueryFile((HDROP)hGlobal, 0, str.GetBuffer(nLength+1), nLength+1);
					str.ReleaseBuffer();
					
					// Open file if Lua script.
					if(str.Right(4) == ".lua")
						AfxGetApp()->OpenDocumentFile(str);

					oleDataObj.Detach();

					return S_FALSE;
				}
			}
		}

		// Release the data source.
		oleDataObj.Detach();
	}

	// If data dropped and accepting data.
	if(bReally && result == S_OK)
	{
		// Set this view active.
		OnActivateView(TRUE, this, NULL);
	}

	return result;
}

int CFEMMedView::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if(CRichEditView::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	// Set default character format.
	CHARFORMAT2 cfDefault;
	cfDefault.cbSize = sizeof(cfDefault);
	cfDefault.dwEffects = CFE_PROTECTED;
	cfDefault.dwMask = CFM_BOLD | CFM_ITALIC | CFM_FACE | CFM_SIZE | CFM_CHARSET | CFM_PROTECTED;
	cfDefault.yHeight = m_lYHeight;
	cfDefault.bCharSet = 0xEE;
	strcpy(cfDefault.szFaceName, _T("Courier New"));

	m_REC.SetDefaultCharFormat(cfDefault);

	// Create listbox font.
	LOGFONT lf;
	memset(&lf, 0, sizeof(LOGFONT));
	lf.lfHeight = 16;
	lf.lfWidth = 8;
	strcpy(lf.lfFaceName, _T("Courier New"));
	m_fontLB.CreateFontIndirect(&lf);

	// Set event mask.
	m_REC.SetEventMask(ENM_CHANGE | ENM_SELCHANGE | ENM_PROTECTED);
	
	return 0;
}

void CFEMMedView::OnChange() 
{
	if(m_bInForcedChange)
		return;

	CHARRANGE crCurSel; 
	m_REC.GetSel(crCurSel);

	if(m_changeType == ctMove && crCurSel.cpMin == crCurSel.cpMax) {
		// cut was canceled, so this is paste operation
		m_changeType = ctPaste;
	}

	switch(m_changeType) {
	case ctReplSel:// old=(x,y) -> cur=(x+len,x+len)
	case ctPaste:  // old=(x,y) -> cur=(x+len,x+len)
		FormatTextLines(m_crOldSel.cpMin, crCurSel.cpMax);
		break;
	case ctDelete: // old=(x,y) -> cur=(x,x)
	case ctBack:   // old=(x,y) -> cur=(x,x), newline del => old=(x,x+1) -> cur=(x-1,x-1)
	case ctCut:	// old=(x,y) -> cur=(x,x)
		FormatTextLines(crCurSel.cpMin, crCurSel.cpMax);
		break;
	case ctUndo:   // old=(?,?) -> cur=(x,y)
		FormatTextLines(crCurSel.cpMin, crCurSel.cpMax);
		break;
	case ctMove:   // old=(x,x+len) -> cur=(y-len,y) | cur=(y,y+len)
		FormatTextLines(crCurSel.cpMin, crCurSel.cpMax);
		if(crCurSel.cpMin > m_crOldSel.cpMin) // move after
			FormatTextLines(m_crOldSel.cpMin, m_crOldSel.cpMin);
		else // move before
			FormatTextLines(m_crOldSel.cpMax, m_crOldSel.cpMax);
		break;
	default:
		FormatAll();
		break;
	}

	//undo action does not call OnProtected, so make it default
	m_changeType = ctUndo;
}

void CFEMMedView::OnProtected(NMHDR* pNMHDR, LRESULT* pResult)
{
	ENPROTECTED* pEP = (ENPROTECTED*)pNMHDR;

	// determine type of change will occur

	switch (pEP->msg) {
	case WM_KEYDOWN:
		switch(pEP->wParam) {
		case VK_DELETE:
			m_changeType = ctDelete;
			break;
		case VK_BACK:
			m_changeType = ctBack;
			break;
		default:
			m_changeType = ctUnknown;
			break;
		}
		break;
	case EM_REPLACESEL:
	case WM_CHAR:
		m_changeType = ctReplSel;
		break;
	case WM_PASTE:
	case EM_PASTESPECIAL:
		m_changeType = (m_changeType == ctCut)?ctMove:ctPaste;
		break;
	case WM_CUT:
		m_changeType = ctCut;
		break;
	case EM_SETCHARFORMAT:
		// Ignore this
		break;
	default:
		m_changeType = ctUnknown;
		break;
	};

	if(pEP->msg != EM_SETCHARFORMAT && m_changeType != ctMove)
		m_crOldSel = pEP->chrg;

	*pResult = FALSE;
}

void CFEMMedView::OnSelChange(NMHDR* pNMHDR, LRESULT* pResult)
{
	SELCHANGE* pSC = (SELCHANGE*)pNMHDR;

	*pResult = 0;
}

LRESULT CFEMMedView::OnSetText(WPARAM wParam, LPARAM lParam)
{
	LRESULT res = Default();

	//FormatAll();
	return res;	
}

void CFEMMedView::FormatAll()
{
	BOOL bSaveModFlag = GetDocument()->IsModified();  // Save modified state.
	FormatTextRange(0, GetTextLength());			  // Otherwise formatting makes
	GetDocument()->SetModifiedFlag(bSaveModFlag);	  // document "dirty".
}

void CFEMMedView::FormatTextLines(const long lLineStart, const long lLineEnd)
{
	int nStart = m_REC.LineIndex(m_REC.LineFromChar(lLineStart));
	int nEnd = m_REC.LineIndex(m_REC.LineFromChar(lLineEnd));
	nEnd += m_REC.LineLength(lLineEnd);

	FormatTextRange(nStart, nEnd);
}

void CFEMMedView::FormatTextRange(const int nStart, const int nEnd)
{
	if(nStart >= nEnd || !m_bOnInitialUpdate)
		return;

	m_bInForcedChange = TRUE;

	CHARRANGE crOldSel;
	CSymbolStyle symSelected;
	const CSymbolStyle symPlain;

	SuspendUndo();
	m_REC.GetSel(crOldSel);

	if(!::g_bSyntaxHighlighting)
	{
		SetFormatRange(nStart, nEnd, symPlain);
		m_REC.SetSel(crOldSel);
		ResumeUndo();
		m_bInForcedChange = FALSE;

		return;
	}

	TCHAR *pBuffer = NULL;
	try {
		m_REC.SetSel(nStart, nEnd);
		pBuffer = new TCHAR[nEnd - nStart + 1];
		long nLen = m_REC.GetSelText(pBuffer);

		ASSERT(nLen <= nEnd - nStart);

		pBuffer[nLen] = 0;

		TCHAR *pStart, *pPtr;
		pStart = pPtr = pBuffer;

		TCHAR* pSymbolStart = NULL;

		while (*pPtr != 0) {
			TCHAR ch = *pPtr;

			if(ch == _T('-') && pPtr[1] == _T('-')) {  // Process comments.
				pSymbolStart = pPtr;
				do {
					ch = *(++pPtr);
				} while (ch != 0 && ch != '\r');
				symSelected = ::g_symComment;				
			} else if(IsStringQuote(ch)) {  // Process strings.
				pSymbolStart = pPtr;
				TCHAR ch1 = ch;
				do {
					ch = *(++pPtr);
				} while (ch != 0 && ch != ch1 && ch != '\r');
				if(ch == ch1) pPtr++;
				symSelected = ::g_symString;
			} else if(_istdigit(ch)) {  // Process numbers.
				pSymbolStart = pPtr;
				_tcstod(pSymbolStart, &pPtr);
				symSelected = ::g_symNumber;
			} else if(_istpunct(ch)) {  // Process punctuation.
				pSymbolStart = pPtr;
				pPtr++;
				symSelected = ::g_symPunctuation;
			} else if(_istalpha(ch) || ch == '_') {  // Process keywords.
				pSymbolStart = pPtr;
				do {
					ch = *(++pPtr);
				} while (_istalnum(ch) || ch == '_');
				*pPtr = 0;
				if(IsLuaKeyword(pSymbolStart)) {
					symSelected = ::g_symKeyword;
				} else {
					int nDummy;
					if(IsFEMMLuaFunction(pSymbolStart, nDummy)) {  // Process functions.
						symSelected = ::g_symFunction;
					} else {
						pSymbolStart = NULL;
					}
				}
				*pPtr = ch;
			} else {
				pPtr++;
			}

			if(pSymbolStart != NULL) {
				ASSERT(pSymbolStart < pPtr);
				SetFormatRange(nStart + pStart - pBuffer, nStart + pSymbolStart - pBuffer, symPlain);
				SetFormatRange(nStart + pSymbolStart - pBuffer, nStart + pPtr - pBuffer, symSelected);
				pStart = pPtr;
				pSymbolStart = 0;
			} else if(*pPtr == 0)
				SetFormatRange(nStart + pStart - pBuffer, nStart + pPtr - pBuffer, symPlain);
		}

	} catch(...){}

	delete [] pBuffer;

	m_REC.SetSel(crOldSel);
	ResumeUndo();
	
	m_bInForcedChange = FALSE;
}

void CFEMMedView::SetFormatRange(const int nStart, const int nEnd, const CSymbolStyle symStyle)
{
	if(nStart >= nEnd)
		return;

	m_REC.SetSel(nStart, nEnd);

	DWORD dwEffects = symStyle.GetBold()?CFE_BOLD:0;
	if(symStyle.GetItalic())
		dwEffects |= CFE_ITALIC;
	if(symStyle.GetUnderline())
		dwEffects |= CFE_UNDERLINE;

	CHARFORMAT2 cfm;
	cfm.cbSize = sizeof(cfm);
	m_REC.GetSelectionCharFormat(cfm);
	
	if((cfm.dwMask & CFM_COLOR)  && cfm.crTextColor == symStyle.GetColor() && 
		(cfm.dwMask & CFM_BOLD & CFM_ITALIC & CFM_UNDERLINE) && 
		(cfm.dwEffects & CFE_BOLD & CFE_ITALIC & CFE_UNDERLINE) == dwEffects)
	{
		return;
	}

	cfm.dwEffects = dwEffects;
	cfm.crTextColor = symStyle.GetColor();
	cfm.dwMask = CFM_BOLD | CFM_ITALIC | CFM_UNDERLINE| CFM_COLOR;

	m_REC.SetSelectionCharFormat(cfm);
}

BOOL CFEMMedView::IsFEMMLuaFunction(LPCTSTR lpszSymbol, int& nIndex)
{
	CString str = lpszSymbol;
	while(str.ReverseFind('_') > 2)  // Remove all but first underscore, e.g. mi'_'.
	{
		str.Delete(str.ReverseFind('_'), 1);
	}
	for(nIndex=0; nIndex<::g_FEMMLuaFunctions.GetSize(); nIndex++)
	{
		CStringArray* pSA = (CStringArray*)::g_FEMMLuaFunctions.GetAt(nIndex);
		if(pSA->GetAt(0).Compare(str) == 0)
		{
			return TRUE;
		}
	}

	return FALSE;
}

BOOL CFEMMedView::IsLuaKeyword(LPCTSTR lpszSymbol, int& nIndex)
{
	for(nIndex=0; nIndex<::g_LuaKeywords.GetSize(); nIndex++)
	{
		CStringArray* pSA = (CStringArray*)::g_LuaKeywords.GetAt(nIndex);
		if(pSA->GetAt(0).Compare(lpszSymbol) == 0)
		{
			return TRUE;
		}
	}

	return FALSE;
}

BOOL CFEMMedView::IsStringQuote(const TCHAR ch)
{
	return (ch == '\'' || ch == '\"');
}

void CFEMMedView::DisableScreenUpdate()
{
	if(m_pTextDoc != NULL)
		m_pTextDoc->Freeze(&m_lFreezeCount);
}

void CFEMMedView::EnableScreenUpdate()
{
	if(m_pTextDoc != NULL)
		m_pTextDoc->Unfreeze(&m_lFreezeCount);
}

void CFEMMedView::ResumeUndo()
{
	m_REC.HideSelection(FALSE, FALSE);
	UnlockWindowUpdate();

	if(m_pTextDoc != NULL)
		m_pTextDoc->Undo(tomResume, NULL);
}

void CFEMMedView::SuspendUndo()
{
	if(m_pTextDoc != NULL)
		m_pTextDoc->Undo(tomSuspend, NULL);

	LockWindowUpdate();
	m_REC.HideSelection(TRUE, FALSE);
}

BOOL CFEMMedView::PreTranslateMessage(MSG* pMsg) 
{
	if(pMsg->hwnd == m_lbListBox.m_hWnd && pMsg->message == WM_CHAR)
	{
		SetFocus();
	}

	if(pMsg->message == WM_KEYDOWN && pMsg->wParam == VK_ESCAPE && m_lbListBox.GetSafeHwnd())
	{
		RemoveListbox();

		return TRUE;  // Don't dispatch message.
	}

	if(pMsg->hwnd == m_lbListBox.m_hWnd && pMsg->message == WM_LBUTTONDBLCLK)
	{
		PostMessage(WM_CHAR, VK_RETURN);

		return TRUE;  // Don't dispatch message.
	}

	if(pMsg->message == WM_MOUSEWHEEL && m_lbListBox.GetSafeHwnd())
	{
		CPoint pt = (CPoint)pMsg->lParam;
		CRect rcLB;
		m_lbListBox.GetWindowRect(&rcLB);

		// If mouse pointer over listbox forward message.
		if(rcLB.PtInRect(pt))
		{
			m_lbListBox.SendMessage(WM_MOUSEWHEEL, pMsg->wParam, pMsg->lParam);
		}
		else
		{
			RemoveListbox();
		}

		return TRUE;  // Don't dispatch message.
	}

	if(pMsg->message == WM_KEYDOWN && pMsg->wParam == VK_INSERT)
	{
		if(pMsg->lParam == 61880)
		{
			m_bLocalOVR = !m_bLocalOVR;
			m_bInToggleOVR = FALSE;
		}
		else
		{
			::g_bGlobalOVR = !::g_bGlobalOVR;
			m_bLocalOVR = !m_bLocalOVR;
		}
	}

	return CRichEditView::PreTranslateMessage(pMsg);
}

// Show popup menu with common commands.
void CFEMMedView::OnRButtonDown(UINT nFlags, CPoint point) 
{
	if(m_lbListBox.GetSafeHwnd())
	{
		RemoveListbox();
	}
	if(m_IntelliTip.GetSafeHwnd())
	{
		m_IntelliTip.Hide();
	}

	// Convert the mouse point to screen coordinates since that is what is used by
	// the TrackPopupMenu() function.
	CPoint ptScreen = point;
	ClientToScreen(&ptScreen);
	CMenu menu;
	menu.CreatePopupMenu();
	
	menu.AppendMenu(MF_STRING, ID_EDIT_UNDO, _T("&Undo"));
	menu.AppendMenu(MF_STRING, ID_EDIT_REDO, _T("&Redo"));
	menu.AppendMenu(MF_SEPARATOR);
	menu.AppendMenu(MF_STRING, ID_EDIT_CUT, _T("Cu&t"));
	menu.AppendMenu(MF_STRING, ID_EDIT_COPY, _T("&Copy"));
	menu.AppendMenu(MF_STRING, ID_EDIT_PASTE, _T("&Paste"));
	menu.AppendMenu(MF_STRING, ID_EDIT_CLEAR, _T("&Delete"));
	menu.AppendMenu(MF_SEPARATOR);
	menu.AppendMenu(MF_STRING, ID_EDIT_SELECT_ALL, _T("Select &All"));
	
	menu.TrackPopupMenu(TPM_LEFTALIGN | TPM_RIGHTBUTTON, ptScreen.x, 
		ptScreen.y, AfxGetMainWnd());

	menu.DestroyMenu();
	
	CRichEditView::OnRButtonDown(nFlags, point);
}

void CFEMMedView::AutoComplete(BOOL bGiveNext)
{
	m_bAutoCompleted = TRUE;

	// Add items.
	if(!m_lbListBox.GetSafeHwnd())
	{
		// Create Auto completion list box.
		CRect p(0, 0, 0, 0);
		m_lbListBox.Create(CBS_DROPDOWNLIST | WS_VSCROLL | WS_CHILD | LBS_STANDARD | LBS_HASSTRINGS, p, this, IDC_LIST);
		m_lbListBox.SetFont(&m_fontLB);

		for(int i=0; i<::g_FEMMLuaFunctions.GetSize(); i++)
		{
			m_lbListBox.AddString(((CStringArray*)::g_FEMMLuaFunctions.GetAt(i))->GetAt(0));
		}
	}
	
	int i=0;
	CString strCurWord;
	if(GetCurWord(strCurWord))
	{
		if(!strCurWord.IsEmpty())
		{
			// Look for nearest.
			int nItems = m_lbListBox.GetCount();
			CString strTemp;

			for( ; i<nItems; i++)
			{
				m_lbListBox.GetText(i, strTemp);
				if(strTemp.CompareNoCase(strCurWord) >= 0)
					break;
			}
		}
	}
	
	// Show listbox if not visible.
	if(!m_lbListBox.IsWindowVisible())
	{
		UpdateLBCoords(strCurWord.GetLength());
		m_lbListBox.ShowWindow(SW_SHOW);
	}
	
	m_lbListBox.SetTopIndex(i);
	m_lbListBox.SetCurSel(i);
		
	return;
}

BOOL CFEMMedView::GetCurWord(CString& strCurWord, const TCHAR ch, BOOL& bWholeWord, BOOL& bFirstWord)
{
	CHARRANGE curSel;
	m_REC.GetSel(curSel);
	
	CString text;
	
	int nLine = m_REC.LineFromChar(curSel.cpMin);
	int nLineLength = m_REC.LineLength(curSel.cpMin);

	if(nLineLength == 0)
	{
		return FALSE;
	}

	char* buffer = new char[nLineLength+4];
	nLineLength = m_REC.GetLine(nLine, buffer, nLineLength);
	
	if(nLineLength > 2 && buffer[nLineLength-1] == '\n' && buffer[nLineLength-2] == '\r')
		buffer[nLineLength-2] = 0;
	else
		buffer[nLineLength] = 0;
	
	int nFrom = curSel.cpMin - m_REC.LineIndex(nLine) - 1;
	int nTo = nFrom;

	if(ch == ' ')  // Find whitespace.
	{
		while(nFrom && !isspace(buffer[nFrom]))
			nFrom--;
	}
	else
	{
		while(nFrom && buffer[nFrom] != ch)
			nFrom--;
	}
	
	if(ch == ' ')
	{
		if(!nFrom && !isspace(buffer[nFrom]))
			nFrom--;
	}
	else
	{
		if(!nFrom && buffer[nFrom] != ch)
			nFrom--;
	}

	bFirstWord = TRUE;
	for(int i=0; i<nFrom; i++)
	{
		if(!isspace(buffer[i]))
		{
			bFirstWord = FALSE;
			break;
		}
	}

	text = buffer + nFrom + 1;
	strCurWord = text.Left(nTo - nFrom);

	if((nTo - nFrom) < text.GetLength())
	{
		TCHAR tchar = text.GetAt(nTo - nFrom);
		bWholeWord = _istspace(tchar);
	}
	else  // End of line.
	{
		bWholeWord = TRUE;
	}

	delete buffer;

	return TRUE;
}

void CFEMMedView::RemoveListbox()
{
	if(m_lbListBox.GetSafeHwnd())
	{
		m_lbListBox.DestroyWindow();
	}
	
	m_bAutoCompleted = FALSE;
}

void CFEMMedView::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	// Override tab key.
	if(::g_bAutoIndentation && nChar == VK_TAB)
	{
		switch(::g_enumIndent)
		{
		case ::FOURSPACES:
			DisableScreenUpdate();
			m_pTextDoc->BeginEditCollection();
			SendMessage(WM_CHAR, ' ');
			SendMessage(WM_CHAR, ' ');
			SendMessage(WM_CHAR, ' ');
			SendMessage(WM_CHAR, ' ');
			m_pTextDoc->EndEditCollection();
			EnableScreenUpdate();
			OnChange();  // Ensures correct syntax highlighting.
			return;
		case ::TWOSPACES:
			DisableScreenUpdate();
			m_pTextDoc->BeginEditCollection();
			SendMessage(WM_CHAR, ' ');
			SendMessage(WM_CHAR, ' ');
			m_pTextDoc->EndEditCollection();
			EnableScreenUpdate();
			OnChange();  // Ensures correct syntax highlighting.
			return;
		}
	}

	if(::g_bIntelliTip)
	{
		if(nChar == '(')  // user has opened a set of brackets
		{
			CString strCurWord;
			GetCurWord(strCurWord);
			int nIndex;
			if(IsFEMMLuaFunction((LPCTSTR)strCurWord, nIndex))
			{
				CStringArray strArrParams;
				CStringArray* pSA = (CStringArray*)::g_FEMMLuaFunctions.GetAt(nIndex);
				for(int i=1; i<pSA->GetSize(); i++)
				{
					strArrParams.Add(pSA->GetAt(i));
				}

				m_IntelliTip.SetFunction(strCurWord, strArrParams);
				m_IntelliTip.SetCurParam(1);

				CPoint pt = GetCaretPos();
				m_IntelliTip.SetCursorPoint(pt);
				m_IntelliTip.Show();
			}
		}
		else if(nChar == ')')
		{
			if(m_IntelliTip.GetSafeHwnd())
				m_IntelliTip.Hide();
		}
		else if(nChar == ',')
		{
			if(m_IntelliTip.GetCurParam() < m_IntelliTip.GetNumParams())
			{
				m_IntelliTip.SetCurParam(m_IntelliTip.GetCurParam() + 1);
			}
		}
	}

	if(!::g_bAutoCompletion)
	{
		CRichEditView::OnChar(nChar, nRepCnt, nFlags);
		AutoIndentationPostChar(nChar);
		return;
	}
	else
	{
		if(nChar == ' ' && GetKeyState(VK_CONTROL) & 0xE000)
		{
			AutoComplete();
			return;
		}
		else if(nChar == VK_RETURN && m_lbListBox.GetSafeHwnd())
		{
			CString strReplace;
			m_lbListBox.GetText(m_lbListBox.GetCurSel(), strReplace);

			DisableScreenUpdate();  // Prevents screen being scrolled.

			CHARRANGE crSel;
			m_REC.GetSel(crSel);

			CString strCurWord;
			if(GetCurWord(strCurWord))
			{
				crSel.cpMin -= strCurWord.GetLength();
				m_REC.SetSel(crSel);
			}

			m_REC.ReplaceSel(strReplace, TRUE);

			EnableScreenUpdate();

			RemoveListbox();
			
			return;
		}
		else if(m_bAutoCompleted)
		{
			if(nChar == VK_BACK)
			{
				CString strCurWord;
				GetCurWord(strCurWord);
				if(!strCurWord.IsEmpty())
				{
					m_bAutoCompleted = FALSE;
					CRichEditView::OnChar(nChar, nRepCnt, nFlags);
					AutoIndentationPostChar(nChar);
					AutoComplete();
				}
				else
				{
					RemoveListbox();
					CRichEditView::OnChar(nChar, nRepCnt, nFlags);
					AutoIndentationPostChar(nChar);
				}
			}
			else if(!isalpha(nChar) && nChar != 95)
			{
				m_bAutoCompleted = FALSE;
				RemoveListbox();
				CRichEditView::OnChar(nChar, nRepCnt, nFlags);
				AutoIndentationPostChar(nChar);
			}
			else
			{
				m_bAutoCompleted = FALSE;
				CRichEditView::OnChar(nChar, nRepCnt, nFlags);
				AutoIndentationPostChar(nChar);				
				AutoComplete();
			}
		}
		else
		{
			m_bAutoCompleted = FALSE;
			CRichEditView::OnChar(nChar, nRepCnt, nFlags);
			AutoIndentationPostChar(nChar);
		}
	}
}

void CFEMMedView::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
 	if(::g_bAutoCompletion && m_lbListBox.GetSafeHwnd())
	{
		// Check for virtual keys if the listbox is visible.
		switch(nChar)
		{
		case(VK_UP):
		case(VK_DOWN):
		case(VK_PRIOR):
		case(VK_NEXT):
		case(VK_HOME):
		case(VK_END):
		case(VK_RETURN):
			{
				m_lbListBox.SendMessage(WM_KEYDOWN, nChar, MAKELPARAM(nRepCnt, nFlags));

				return;
			}
		case(VK_TAB):
			{
				RemoveListbox();

				return;
			} 
		case(VK_RIGHT):
			{
				CHARRANGE curSel;
				m_REC.GetSel(curSel);
				m_REC.SetSel(curSel.cpMin+1, curSel.cpMax);
				AutoComplete();

				return;
			}
		case(VK_LEFT):
			{
				CString strCurWord;
				GetCurWord(strCurWord);
				
				if(!strCurWord.IsEmpty())
				{
					CHARRANGE curSel;
					m_REC.GetSel(curSel);
					m_REC.SetSel(curSel.cpMin-1, curSel.cpMax);
					AutoComplete();
				}
				else
				{
					m_REC.ReplaceSel("");
					RemoveListbox();
				}

				return;
			}
		case(VK_ESCAPE):  // Not being called so added to PreTranslateMessage().
			{
				RemoveListbox();
				m_REC.ReplaceSel("");
				
				return;
			}
		}
	}

	if(::g_bIntelliTip)
	{
		if((nChar == VK_UP || nChar == VK_DOWN || nChar == VK_RETURN || nChar == VK_HOME
			|| nChar == VK_END))
		{
			m_IntelliTip.Hide();
		}
		else if(nChar == VK_BACK && m_IntelliTip.IsShown())
		{
			CHARRANGE curSel;
			m_REC.GetSel(curSel);
			if(curSel.cpMin > 0)
			{
				SuspendUndo();

				m_REC.SetSel(curSel.cpMin - 1, curSel.cpMin);
				CString strTemp = m_REC.GetSelText();
				m_REC.SetSel(curSel);  // set selection back to original

				ResumeUndo();
		
				if(strTemp == "(")
				{
					m_IntelliTip.Hide();
				}
			}
		}
	}
	
	AutoIndentationPreChar(nChar);
	CRichEditView::OnKeyDown(nChar, nRepCnt, nFlags);

	if(::g_bIntelliTip)
	{
		if((nChar == VK_BACK  || nChar == VK_DELETE || nChar == VK_LEFT || nChar == VK_RIGHT)
			&& m_IntelliTip.IsShown())
		{
			CString strCurWord;
			GetCurWord(strCurWord, '(');
			strCurWord = strCurWord.Right(strCurWord.GetLength() - strCurWord.ReverseFind('('));
			m_IntelliTip.SetCurParam(strCurWord.Remove(',') + 1);
		}
	}
}

void CFEMMedView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	RemoveListbox();
	if(m_IntelliTip.GetSafeHwnd())
	{
		m_IntelliTip.Hide();
	}
	
	CRichEditView::OnLButtonDown(nFlags, point);
}

int CFEMMedView::GetLBLongestString()
{
	int nReturn = 0;

	for(int i=0; i<m_lbListBox.GetCount(); i++)
	{
		if(m_lbListBox.GetTextLen(i) > nReturn)
		{
			nReturn = m_lbListBox.GetTextLen(i);
		}
	}

	return nReturn;
}

// Must be called after listbox has been populated with CListBox::AddString().
void CFEMMedView::UpdateLBCoords(int nCurWordLen)
{
	CRect rcLB, rcWnd;
	m_lbListBox.GetWindowRect(&rcLB);
	ScreenToClient(&rcLB);
	GetWindowRect(&rcWnd);
	ScreenToClient(&rcWnd);

	LOGFONT lf;
	m_fontLB.GetLogFont(&lf);

	long lHeight = m_lbListBox.GetItemHeight(0) * min(m_lbListBox.GetCount() + 1, 10);
	long lWidth = GetLBLongestString() * lf.lfWidth + 30;  // 30px to allow for scrollbar.

	CHARRANGE crSel;
	m_REC.GetSel(crSel);

	rcLB.top  = m_REC.GetCharPos(crSel.cpMin - nCurWordLen).y + (m_lYHeight/10);
	rcLB.left = m_REC.GetCharPos(crSel.cpMin - nCurWordLen).x;

	// Top and bottom.
	if(rcLB.top + lHeight > rcWnd.bottom)
		rcLB.top = rcLB.top - (m_lYHeight/20) - lHeight;
	
	rcLB.bottom = rcLB.top + lHeight;

	// Left and right;
	if(rcLB.left + lWidth > rcWnd.right)
		rcLB.left = rcLB.left - (rcLB.left + lWidth - rcWnd.right) - 10;

	rcLB.right = rcLB.left + lWidth;

	m_lbListBox.MoveWindow(&rcLB);
}

void CFEMMedView::OnUpdateEditRedo(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(SendMessage(EM_CANREDO, 0, 0));
}

void CFEMMedView::OnEditRedo() 
{
	SendMessage(EM_REDO, 0, 0);
}

LRESULT CFEMMedView::OnFindDialogMessage(WPARAM wParam, LPARAM lParam)
{
    // Get current selection.
	CHARRANGE crSel;
	m_REC.GetSel(crSel);

	// If Find/Replace dialog not created,
	// has message come from CFEMMedView::OnEditFindNext()?
	if(m_pFindDialog == NULL)
	{
		if(lParam == FR_FINDNEXT)
		{
			if(FindString(crSel))
			{
				m_REC.SetSel(crSel);
			}
		}
	}
	else
	{
		// Read data from Find/Replace dialog.
		::g_strFind = m_pFindDialog->GetFindString();
		::g_strReplace = m_pFindDialog->GetReplaceString();
		::g_bMatchCase = m_pFindDialog->MatchCase();
		::g_bMatchWholeWord = m_pFindDialog->MatchWholeWord();

		// If the FR_DIALOGTERM flag is set,
		// invalidate the handle identifying the dialog box.
		if(m_pFindDialog->IsTerminating())
		{
			m_pFindDialog = NULL;
			return 0;
		}

		// Else if the FR_FINDNEXT flag is set,
		// search for the requested string.
		else if(m_pFindDialog->FindNext())
		{
			if(FindString(crSel))
			{
				m_REC.SetSel(crSel);
			}
			else
			{
				MessageBox("FEMMed cannot find \"" + ::g_strFind + "\".", "FEMMed", MB_OK \
					| MB_ICONWARNING);
				if(m_pFindDialog != NULL)
					m_pFindDialog->SetFocus();
			}
		}

		// Else if the FR_REPLACEALL flag is set,
		// replace all instances of ::g_strFind with ::g_strReplace.
		else if(m_pFindDialog->ReplaceAll())
		{
			BOOL nFound = FALSE;
			int nReplaced = 0;
			do
			{
				nFound = FindString(crSel);
				if(nFound)
				{
					m_REC.SetSel(crSel);
					m_REC.ReplaceSel(::g_strReplace, TRUE);
					crSel.cpMin += ::g_strReplace.GetLength();
					crSel.cpMax = crSel.cpMin;

					nReplaced++;
				}
				else
				{
					CString strMes;
					strMes.Format("Replaced %i occurrence(s) of '%s' with '%s'.", nReplaced, \
						::g_strFind, ::g_strReplace);
					MessageBox(strMes, "FEMMed", MB_OK | MB_ICONASTERISK);
					if(m_pFindDialog != NULL)
						m_pFindDialog->SetFocus();
				}
			} while(nFound);
		}

		// Else if the FR_REPLACE flag is set,
		// replace the current selection if it matches ::g_strFind.
		else if(m_pFindDialog->ReplaceCurrent())
		{
			CString strCurSel = m_REC.GetSelText();

			int nComp;

			if(::g_bMatchCase)
				nComp = strCurSel.Compare(::g_strFind);
			else
				nComp = strCurSel.CompareNoCase(::g_strFind);

			if(nComp == 0)
			{
				m_REC.ReplaceSel(::g_strReplace, TRUE);
				m_REC.SetSel(crSel.cpMin, crSel.cpMin + ::g_strReplace.GetLength());
			}
		}
	}

    return 0;
}

BOOL CFEMMedView::FindString(CHARRANGE& crSel)
{
	BOOL bReturn = FALSE;

	DWORD dwFlags = FR_DOWN;
	if(::g_bMatchCase)
		dwFlags |= FR_MATCHCASE;
	if(::g_bMatchWholeWord)
		dwFlags |= FR_WHOLEWORD;
	
	FINDTEXTEX ft;
	ft.chrg.cpMin = crSel.cpMax;
	ft.chrg.cpMax = -1;  // End.
	ft.lpstrText = (LPSTR)(LPCTSTR)::g_strFind;

	if(m_REC.FindText(dwFlags, &ft) != -1)
	{
		crSel = ft.chrgText;

		bReturn = TRUE;
	}
	else
	{
		// Reached end of file. Keep searching from start?
		if(MessageBox("Reached the end of file.\nContinue from start?", "FEMMed - Find", \
			MB_ICONWARNING | MB_YESNO) == IDYES)
		{
			ft.chrg.cpMin = 0;  // Start.
			ft.chrg.cpMax = crSel.cpMax;
	
			if(m_REC.FindText(dwFlags, &ft) != -1)
			{
				crSel = ft.chrgText;

				bReturn = TRUE;
			}
		}
		if(m_pFindDialog != NULL)
		{
			m_pFindDialog->SetFocus();
		}
	}

	if(bReturn && m_pFindDialog != NULL)
	{
		// Move dialog if necessary. Work in screen coordinates.
		CPoint ptStart, ptEnd;
		ptStart = m_REC.GetCharPos(crSel.cpMin);
		ClientToScreen(&ptStart);
		ptEnd = m_REC.GetCharPos(crSel.cpMax);
		ClientToScreen(&ptEnd);

		CRect rcSel;
		rcSel.top = ptStart.y - m_lYHeight/50;
		rcSel.bottom = ptEnd.y + m_lYHeight/10;
		rcSel.left = ptStart.x;
		rcSel.right = ptEnd.x;
		
		CRect rcWnd, rcView, rcIntersect;
		m_pFindDialog->GetWindowRect(rcWnd);  // Already screen coordinates.
		GetClientRect(rcView);
		ClientToScreen(rcView);

		if(rcIntersect.IntersectRect(rcWnd, rcSel))  // Dialog overlaps selection.
		{
			int nMoveY = rcSel.bottom - rcWnd.top;
			int nHeight = rcWnd.Height();

			// Ensure dialog doesn't get moved off the bottom of the view.
			if(rcWnd.bottom + nMoveY <= rcView.bottom)
				rcWnd.bottom += nMoveY;
			else
				rcWnd.bottom = rcSel.top;

			rcWnd.top = rcWnd.bottom - nHeight;

			m_pFindDialog->MoveWindow(rcWnd);
		}
	}

	return bReturn;
}

void CFEMMedView::OnEditFind() 
{
	if(m_pFindDialog == NULL)
	{
		CHARRANGE crSel;
		m_REC.GetSel(crSel);
		if(crSel.cpMin != crSel.cpMax)
			::g_strFind = m_REC.GetSelText();

		DWORD dwFlags = FR_NOUPDOWN | FR_HIDEUPDOWN;
		if(::g_bMatchCase)
			dwFlags |= FR_MATCHCASE;
		if(::g_bMatchWholeWord)
			dwFlags |= FR_WHOLEWORD;

		m_pFindDialog = new CFindReplaceDialog();
		m_pFindDialog->Create(TRUE, ::g_strFind, ::g_strReplace, dwFlags, this);
	}
	else
	{
		m_pFindDialog->SetFocus();
	}
}

void CFEMMedView::OnEditReplace() 
{
	if(m_pFindDialog == NULL)
	{		
		CHARRANGE crSel;
		m_REC.GetSel(crSel);
		if(crSel.cpMin != crSel.cpMax)
			::g_strFind = m_REC.GetSelText();

		DWORD dwFlags = FR_NOUPDOWN | FR_HIDEUPDOWN;
		if(::g_bMatchCase)
			dwFlags |= FR_MATCHCASE;
		if(::g_bMatchWholeWord)
			dwFlags |= FR_WHOLEWORD;

		m_pFindDialog = new CFindReplaceDialog();
		m_pFindDialog->Create(FALSE, ::g_strFind, ::g_strReplace, dwFlags, this);
	}
	else
	{
		m_pFindDialog->SetFocus();
	}
}

void CFEMMedView::OnEditFindNext() 
{
	SendMessage(m_FindDialogMessage, 0, FR_FINDNEXT);
}

void CFEMMedView::OnUpdateEditFindNext(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(!::g_strFind.IsEmpty() && GetTextLength());
}

void CFEMMedView::AutoIndentationPreChar(const UINT nChar)
{
	if(::g_bAutoIndentation)
	{
		if(nChar == VK_RETURN)
		{
			switch(::g_enumIndent)
			{
			case ::TAB:
				m_strIndent = "\t";
				break;
			case ::FOURSPACES:
				m_strIndent = "    ";
				break;
			case ::TWOSPACES:
				m_strIndent = "  ";
				break;
			}

			int nLine = m_REC.LineFromChar(-1);
			int nStart = m_REC.LineIndex(nLine);
			int nLineLen = m_REC.LineLength(-1);
			if(nLineLen < 4)  // Must be at least 4 bytes for CRichEditCtrl::GetLine().
				nLineLen = 4;
			
			CHARRANGE crSel;
			m_REC.GetSel(crSel);

			CString strLine;
			nLineLen = m_REC.GetLine(nLine, strLine.GetBuffer(nLineLen), nLineLen);
			strLine.ReleaseBuffer(nLineLen);

			strLine = strLine.Left(crSel.cpMin - nStart);
			strLine += "\r";

			m_nIndent = strLine.GetLength();
			strLine.TrimLeft(m_strIndent);
			m_nIndent -= strLine.GetLength();

			m_nIndent /= m_strIndent.GetLength();

			BOOL bIndent = TRUE;
			for(int i=0; i<::g_LuaKeywords.GetSize(); i++)
			{
				CStringArray* pSA = (CStringArray*)::g_LuaKeywords.GetAt(i);
				for(int j=1; j<pSA->GetSize(); j++)
				{
					if(pSA->GetAt(j).CompareNoCase("outdent") == 0)
					{
						if(strLine.Find(" " + pSA->GetAt(0) + " ") >= 0
							|| strLine.Find(pSA->GetAt(0) + " ") > 0
							|| strLine.Find(" " + pSA->GetAt(0) + "\r") >= 0)
						{
							bIndent = FALSE;
							break;
						}
					}
				}
				if(!bIndent)
					break;
			}

			if(bIndent)
			{
				strLine = strLine.Left(strLine.FindOneOf(" \r"));
				int nFoundIndex;
				if(IsLuaKeyword((LPCTSTR)strLine, nFoundIndex))
				{
					CStringArray* pSA = (CStringArray*)::g_LuaKeywords.GetAt(nFoundIndex);
					for(i=1; i<pSA->GetSize(); i++)
					{
						if(pSA->GetAt(i).CompareNoCase("indent") == 0)
						{
							m_nIndent++;
							break;
						}
					}
				}
			}

			DisableScreenUpdate();
		}
	}
}

void CFEMMedView::AutoIndentationPostChar(const UINT nChar)
{
	if(::g_bAutoIndentation)
	{
		if(nChar == VK_RETURN && !m_lbListBox.GetSafeHwnd())
		{
			CString strIndent;
			int nIndent = m_nIndent;
			for( ; nIndent>0 ; nIndent--)
				strIndent += m_strIndent;

			m_REC.ReplaceSel(strIndent, TRUE);

			m_bOutdent = TRUE;

			EnableScreenUpdate();
		}

		if(!m_lbListBox.GetSafeHwnd() && isalpha(nChar) && m_bOutdent)
		{
			if(m_REC.LineLength(-1) < m_strIndent.GetLength())  // Can't outdent.
				return;

			CHARRANGE crOldSel;
			m_REC.GetSel(crOldSel);

			CString strCurWord;
			BOOL bWholeWord, bFirstWord;
			if(GetCurWord(strCurWord, ' ', bWholeWord, bFirstWord))
			{
				if(!bWholeWord || !bFirstWord)
					return;

				int nFoundIndex;
				if(IsLuaKeyword((LPCTSTR)strCurWord, nFoundIndex))
				{
					CStringArray* pSA = (CStringArray*)::g_LuaKeywords.GetAt(nFoundIndex);
					for(int i=1; i<pSA->GetSize(); i++)
					{
						if(pSA->GetAt(i).CompareNoCase("outdent") == 0)
						{
							DisableScreenUpdate();
							m_pTextDoc->BeginEditCollection();

							int nLine = m_REC.LineFromChar(-1);
							int nStart = m_REC.LineIndex(nLine);
							CHARRANGE crSel;

							crSel.cpMin = nStart;
							crSel.cpMax = nStart + m_strIndent.GetLength();
							m_REC.SetSel(crSel);
							strCurWord = m_REC.GetSelText();

							if(strCurWord.Find(m_strIndent) == 0)
							{
								m_REC.ReplaceSel("", TRUE);
								crOldSel.cpMin -= m_strIndent.GetLength();
								crOldSel.cpMax -= m_strIndent.GetLength();

								m_bOutdent = FALSE;
							}

							m_REC.SetSel(crOldSel);

							m_pTextDoc->EndEditCollection();
							EnableScreenUpdate();

							break;
						}
					}
				}
			}
		}
	}
}

void CFEMMedView::OnEditUndo() 
{
	if(m_REC.CanUndo())
	{
		m_REC.Undo();
	}

	if(!m_REC.CanUndo())  // If the undo buffer is empty, reset the IsModified() flag.
	{
		GetDocument()->SetModifiedFlag(FALSE);
	}
}

void CFEMMedView::OnActivateView(BOOL bActivate, CView* pActivateView, CView* pDeactiveView) 
{
	if(bActivate)
	{
		// Ensure correct OVR/INS state.
		if((m_bLocalOVR != ::g_bGlobalOVR) && !m_bInToggleOVR)
		{
			WPARAM wParam = VK_INSERT;
			LPARAM lParam = 61880;  // Arbitrary number to look for in
									// PreTranslateMessage().

			PostMessage(WM_KEYDOWN, wParam, lParam);
			PostMessage(WM_KEYUP  , wParam, lParam);

			m_bInToggleOVR = TRUE;
		}
	}

	else
	{
		// Hide listbox, IntelliTip and Find/Replace dialog.
		if(m_lbListBox.GetSafeHwnd())
		{
			RemoveListbox();
		}
		if(m_IntelliTip.GetSafeHwnd())
		{
			m_IntelliTip.Hide();
		}
		if(m_pFindDialog != NULL)
		{
			m_pFindDialog->DestroyWindow();
			m_pFindDialog = NULL;
		}
	}
	
	CRichEditView::OnActivateView(bActivate, pActivateView, pDeactiveView);
}

void CFEMMedView::OnUpdatePositionIndicator(CCmdUI* pCmdUI)
{
	// Get the character index of the line with the caret
	int nLineIndex = m_REC.LineIndex(-1);

	// Get the line number
	int nLine = m_REC.LineFromChar(nLineIndex) + 1;

	// Get the column
	long lStart, lEnd;
	m_REC.GetSel(lStart, lEnd);
	int nCol = lEnd > nLineIndex ? lEnd - nLineIndex + 1 : 1;

	CString str;
	str.Format(_T("Ln %d, Col %d"), nLine, nCol);

	pCmdUI->SetText(str);

	pCmdUI->Enable(TRUE);
}
