// FEMMedView.h : interface of the CFEMMedView class
//
/////////////////////////////////////////////////////////////////////////////

#if !defined(AFX_FEMMEDVIEW_H__FC1E68F5_E5B4_4E9E_A3D9_5EA6D4AEAFF1__INCLUDED_)
#define AFX_FEMMEDVIEW_H__FC1E68F5_E5B4_4E9E_A3D9_5EA6D4AEAFF1__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "TOM.h"
#include <Richole.h>

#include "SymbolStyle.h"
#include "Intellitip.h"

class CFEMMedCntrItem;

class CFEMMedView : public CRichEditView
{
protected: // create from serialization only
	CFEMMedView();
	DECLARE_DYNCREATE(CFEMMedView)

// Attributes
public:
	CFEMMedDoc* GetDocument();
private:
	CRichEditCtrl& m_REC;
	ITextDocument* m_pTextDoc;
	CFont m_fontLB;  // Font for listbox.
	BOOL m_bLocalOVR;  // OVR state.
	BOOL m_bInToggleOVR;
	BOOL m_bOnInitialUpdate;
	long m_lYHeight;  // Height of characters in CHARFORMAT2.
	long m_lFreezeCount;
	void DisableScreenUpdate();
	void EnableScreenUpdate();
	void ResumeUndo();
	void SuspendUndo();

// Find/Replace members.
protected:
	static const UINT m_FindDialogMessage;
	CFindReplaceDialog* m_pFindDialog;

	BOOL FindString(CHARRANGE&);

// Auto completion.
public:
	void RemoveListbox();
private:
	void AutoComplete(BOOL bGiveNext = TRUE);
	BOOL GetCurWord(CString& strCurWord) { BOOL b1,b2; return GetCurWord(strCurWord,' ',b1,b2); }
	BOOL GetCurWord(CString& strCurWord, const TCHAR ch) { BOOL b1,b2;return GetCurWord(strCurWord,ch,b1,b2); }
	BOOL GetCurWord(CString& strCurWord, const TCHAR ch, BOOL& bWholeWord, BOOL& bFirstWord);
	int GetLBLongestString();
	void UpdateLBCoords(int nCurWordLen);
	CListBox m_lbListBox;
	BOOL m_bAutoCompleted;

// Auto indentation.
	void AutoIndentationPreChar(const UINT nChar);
	void AutoIndentationPostChar(const UINT nChar);
	CString m_strIndent;
	int m_nIndent;
	BOOL m_bOutdent;

// IntelliTip.
public:
	CIntelliTip m_IntelliTip;

// Syntax highlighting.
public:
	void FormatAll();
private:
	void FormatTextLines(const long lLineStart, const long lLineEnd);
	void FormatTextRange(const int nStart, const int nEnd);
	void SetFormatRange(const int nStart, const int nEnd, const CSymbolStyle symStyle);
	BOOL IsStringQuote(const TCHAR ch);
	BOOL IsFEMMLuaFunction(LPCTSTR lpszSymbol) { int n; return IsFEMMLuaFunction(lpszSymbol, n); }
	BOOL IsFEMMLuaFunction(LPCTSTR lpszSymbol, int& nFoundIndex);
	BOOL IsLuaKeyword(LPCTSTR lpszSymbol) { int nDummy; return IsLuaKeyword(lpszSymbol, nDummy); }
	BOOL IsLuaKeyword(LPCTSTR lpszSymbol, int& nFoundIndex);
	
	enum ChangeType {ctUndo, ctUnknown, ctReplSel, ctDelete, ctBack, ctCut, ctPaste, ctMove};
	ChangeType m_changeType;

	BOOL m_bInForcedChange;
	CHARRANGE m_crOldSel;

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CFEMMedView)
	public:
	virtual BOOL PreTranslateMessage(MSG* pMsg);
	protected:
	virtual void OnInitialUpdate(); // called first time after construct
	virtual BOOL OnPreparePrinting(CPrintInfo* pInfo);
	virtual void OnPrint(CDC* pDC, CPrintInfo* pInfo);
	virtual void OnActivateView(BOOL bActivate, CView* pActivateView, CView* pDeactiveView);
	//}}AFX_VIRTUAL
	virtual HRESULT QueryAcceptData(LPDATAOBJECT, CLIPFORMAT*, DWORD,
		BOOL, HGLOBAL);

// Implementation
public:
	virtual ~CFEMMedView();
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

protected:

// Generated message map functions
protected:
	//{{AFX_MSG(CFEMMedView)
	afx_msg void OnDestroy();
	afx_msg int OnCreate(LPCREATESTRUCT lpCreateStruct);
	afx_msg void OnChange();
	afx_msg void OnEditPaste();
	afx_msg void OnUpdateEditPaste(CCmdUI* pCmdUI);
	afx_msg void OnRButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnChar(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnLButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnEditUndo();
	afx_msg void OnEditRedo();
	afx_msg void OnUpdateEditRedo(CCmdUI* pCmdUI);
	afx_msg void OnEditFind();
	afx_msg void OnEditReplace();
	afx_msg void OnEditFindNext();
	afx_msg void OnUpdateEditFindNext(CCmdUI* pCmdUI);
	afx_msg void OnUpdatePositionIndicator(CCmdUI* pCmdUI);
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
	afx_msg LRESULT OnSetText(WPARAM wParam, LPARAM lParam);
	afx_msg void OnProtected(NMHDR*, LRESULT* pResult);
	afx_msg void OnSelChange(NMHDR*, LRESULT* pResult);
	afx_msg LRESULT OnFindDialogMessage(WPARAM wParam, LPARAM lParam);
};

#ifndef _DEBUG  // debug version in FEMMedView.cpp
inline CFEMMedDoc* CFEMMedView::GetDocument()
   { return (CFEMMedDoc*)m_pDocument; }
#endif

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_FEMMEDVIEW_H__FC1E68F5_E5B4_4E9E_A3D9_5EA6D4AEAFF1__INCLUDED_)
